diff --git a/.isort.cfg b/.isort.cfg
index 4a48dba8d2bcd070f0644cace52089cc21ab8b00..6b83d3b7f8338af28d1685dbd994088e0b2ce666 100644
--- a/.isort.cfg
+++ b/.isort.cfg
@@ -8,3 +8,4 @@ known_first_party=finn
 sections=FUTURE,STDLIB,COMPAT,TEST,THIRDPARTY,FIRSTPARTY,LOCALFOLDER
 default_section=THIRDPARTY
 multi_line_output=3
+include_trailing_comma=True
diff --git a/src/finn/analysis/fpgadataflow/__init__.py b/src/finn/analysis/fpgadataflow/__init__.py
new file mode 100644
index 0000000000000000000000000000000000000000..e69de29bb2d1d6434b8b29ae775ad8c2e48c5391
diff --git a/src/finn/analysis/fpgadataflow/hls_synth_res_estimation.py b/src/finn/analysis/fpgadataflow/hls_synth_res_estimation.py
new file mode 100644
index 0000000000000000000000000000000000000000..d7485b924e7ddb7f59fbc9e316df4d74e50218bc
--- /dev/null
+++ b/src/finn/analysis/fpgadataflow/hls_synth_res_estimation.py
@@ -0,0 +1,48 @@
+import os
+import xml.etree.ElementTree as ET
+
+import finn.core.utils as util
+import finn.custom_op.registry as registry
+
+
+def hls_synth_res_estimation(model):
+    """Extracts the results from the vivado synthesis.
+    Returns {node name : resource estimation}"""
+
+    res_dict = {}
+    for node in model.graph.node:
+        if node.domain == "finn":
+            backend_attribute = util.get_by_name(node.attribute, "backend")
+            if backend_attribute is None:
+                continue
+            backend_value = backend_attribute.s.decode("UTF-8")
+            if backend_value == "fpgadataflow":
+                op_type = node.op_type
+                inst = registry.custom_op[op_type](node)
+                code_gen_dir = inst.get_nodeattr("code_gen_dir_ipgen")
+                if code_gen_dir == "":
+                    raise Exception(
+                        """Please run "CodeGen_ipgen" transformation and
+                            "HLSSynth_IPGen" first to generate the report files"""
+                    )
+                else:
+                    xmlfile = "{}/project_{}/sol1/syn/report/{}_csynth.xml".format(
+                        code_gen_dir, node.name, node.name
+                    )
+
+                    if os.path.isfile(xmlfile):
+                        res_dict[node.name] = []
+                        tree = ET.parse(xmlfile)
+                        root = tree.getroot()
+                        for item in root.findall("AreaEstimates/Resources"):
+                            for child in item:
+                                res_dict[node.name].append(
+                                    ["{} : {}".format(child.tag, child.text)]
+                                )
+                    else:
+                        raise Exception(
+                            """Please run "HLSSynth_IPGen" first
+                                to generate the report files"""
+                        )
+
+    return res_dict
diff --git a/src/finn/analysis/fpgadataflow/res_estimation.py b/src/finn/analysis/fpgadataflow/res_estimation.py
new file mode 100644
index 0000000000000000000000000000000000000000..1693e413dcf0f2bee2587984c4f1db9de8a9cb68
--- /dev/null
+++ b/src/finn/analysis/fpgadataflow/res_estimation.py
@@ -0,0 +1,21 @@
+import finn.core.utils as util
+import finn.custom_op.registry as registry
+
+
+def res_estimation(model):
+    """Estimates the resources needed for the given model.
+    Returns {node name : resource estimation}"""
+
+    res_dict = {}
+    for node in model.graph.node:
+        if node.domain == "finn":
+            backend_attribute = util.get_by_name(node.attribute, "backend")
+            if backend_attribute is None:
+                continue
+            backend_value = backend_attribute.s.decode("UTF-8")
+            if backend_value == "fpgadataflow":
+                op_type = node.op_type
+                inst = registry.custom_op[op_type](node)
+                res_dict[node.name] = inst.node_res_estimation()
+
+    return res_dict
diff --git a/src/finn/backend/fpgadataflow/utils.py b/src/finn/backend/fpgadataflow/utils.py
index 0f3049ec70050657c4a648fe8b51a2d16691bed0..257c2b0988b6dd8ed724c37cef120b1c4f0da473 100644
--- a/src/finn/backend/fpgadataflow/utils.py
+++ b/src/finn/backend/fpgadataflow/utils.py
@@ -3,7 +3,10 @@ import sys
 import numpy as np
 
 from finn.core.datatype import DataType
-from finn.core.utils import pack_innermost_dim_as_hex_string
+from finn.core.utils import (
+    pack_innermost_dim_as_hex_string,
+    unpack_innermost_dim_from_hex_string,
+)
 
 
 def numpy_to_hls_code(
@@ -56,3 +59,38 @@ def numpy_to_hls_code(
     else:
         ret = ret + " = \n" + strarr + ";"
     return ret
+
+
+def npy_to_rtlsim_input(input_file, input_dtype, pad_to_nbits):
+    """Convert the multidimensional NumPy array of integers (stored as floats)
+    from input_file into a flattened sequence of Python arbitrary-precision
+    integers, packing the innermost dimension. See
+    finn.core.utils.pack_innermost_dim_as_hex_string() for more info on how the
+    packing works."""
+
+    inp = np.load(input_file)
+    ishape = inp.shape
+    inp = inp.flatten()
+    inp_rev = []
+    for i in range(len(inp)):
+        inp_rev.append(inp[-1])
+        inp = inp[:-1]
+    inp_rev = np.asarray(inp_rev, dtype=np.float32).reshape(ishape)
+    packed_data = pack_innermost_dim_as_hex_string(inp_rev, input_dtype, pad_to_nbits)
+    packed_data = packed_data.flatten()
+    packed_data = [int(x[2:], 16) for x in packed_data]
+    packed_data.reverse()
+    return packed_data
+
+
+def rtlsim_output_to_npy(output, path, dtype, shape, packedBits, targetBits):
+    """Convert a flattened sequence of Python arbitrary-precision integers
+    output into a NumPy array, saved as npy file at path. Each arbitrary-precision
+    integer is assumed to be a packed array of targetBits-bit elements, which
+    will be unpacked as the innermost dimension of the NumPy array."""
+
+    output = [hex(int(x)) for x in output]
+    out_array = unpack_innermost_dim_from_hex_string(
+        output, dtype, shape, packedBits, targetBits, True
+    )
+    np.save(path, out_array)
diff --git a/src/finn/core/utils.py b/src/finn/core/utils.py
index 0a6e5718d86edb861046b2753b4ab8a4b594c5cc..eb96eb064aabb3f10a42f8ff1686f86092519a18 100644
--- a/src/finn/core/utils.py
+++ b/src/finn/core/utils.py
@@ -1,7 +1,7 @@
+import os
 import random
 import string
 import subprocess
-import os
 
 import numpy as np
 import onnx
@@ -111,6 +111,49 @@ def pack_innermost_dim_as_hex_string(ndarray, dtype, pad_to_nbits):
     return np.apply_along_axis(fun, ndarray.ndim - 1, ndarray)
 
 
+def unpack_innermost_dim_from_hex_string(
+    data, dtype, shape, packedBits, targetBits, rtlsim=False
+):
+    # function expects flattens array and returns an array in the desired shape
+    outer_dim_elems = 1
+    for dim in range(len(shape) - 1):
+        outer_dim_elems = outer_dim_elems * shape[dim]
+    inner_dim_elems = shape[-1]
+
+    array = []
+    for outer_elem in range(outer_dim_elems):
+        ar_list = []
+        ar_elem = data[0]
+        data.pop(0)
+        ar_elem = ar_elem.split("x")
+        ar_elem_bin = bin(int(ar_elem[1], 16))[2:].zfill(packedBits)
+        ar_elem_bin = [int(x) for x in ar_elem_bin]
+
+        ar_elem_bin.reverse()
+        for i in range(inner_dim_elems):
+            upper_limit = (i + 1) * targetBits
+            lower_limit = i * targetBits
+            elem = ar_elem_bin[lower_limit:upper_limit]
+            elem.reverse()
+            elem_str = "".join(map(str, elem))
+            ar_list.append(int(elem_str, 2))
+        # reverse inner dimension back to "normal" positions
+        if rtlsim is False:
+            ar_list.reverse()
+        else:
+            # interpret output values correctly by flattening and adjusting the output
+            if dtype == DataType.BIPOLAR:
+                ar_list = [2 * x - 1 for x in ar_list]
+            # pyverilator interprets int2 as uint2, so output has to be corrected
+            elif dtype == DataType.INT2 or dtype == DataType.INT32:
+                mask = 2 ** (dtype.bitwidth() - 1)
+                ar_list = [-(x & mask) + (x & ~mask) for x in ar_list]
+
+        array.append(ar_list)
+    array = np.asarray(array, dtype=np.float32).reshape(shape)
+    return array
+
+
 def interleave_matrix_outer_dim_from_partitions(matrix, n_partitions):
     if type(matrix) != np.ndarray or matrix.dtype != np.float32:
         # try to convert to a float numpy array (container dtype is float)
@@ -202,7 +245,7 @@ def calculate_signed_dot_prod_range(dt_a, dt_b, len):
     types dt_a and dt_b of len elements can take."""
     assert dt_a.signed() and dt_b.signed()
     min_prod = 2 ** 30
-    max_prod = -2 ** 30
+    max_prod = -(2 ** 30)
     for a_val in [dt_a.min(), dt_a.max()]:
         for b_val in [dt_b.min(), dt_b.max()]:
             prod = a_val * b_val * len
@@ -250,12 +293,13 @@ class CppBuilder:
         process_compile = subprocess.Popen(bash_command, stdout=subprocess.PIPE)
         process_compile.communicate()
 
+
 class IPGenBuilder:
     def __init__(self):
         self.tcl_script = ""
         self.ipgen_path = ""
         self.code_gen_dir = ""
-        self.ipgen_script=""
+        self.ipgen_script = ""
 
     def append_tcl(self, tcl_script):
         self.tcl_script = tcl_script
@@ -276,4 +320,3 @@ class IPGenBuilder:
         bash_command = ["bash", self.ipgen_script]
         process_compile = subprocess.Popen(bash_command, stdout=subprocess.PIPE)
         process_compile.communicate()
-        
diff --git a/src/finn/custom_op/fpgadataflow/__init__.py b/src/finn/custom_op/fpgadataflow/__init__.py
index 3f64b8940bf0e9ffdd03086735818c49d96bf1c2..b2db174d9da294e68bd12026bf9cc6f849cf5f23 100644
--- a/src/finn/custom_op/fpgadataflow/__init__.py
+++ b/src/finn/custom_op/fpgadataflow/__init__.py
@@ -5,7 +5,6 @@ import subprocess
 from finn.custom_op import CustomOp
 from finn.core.utils import CppBuilder, IPGenBuilder
 import finn.custom_op.fpgadataflow.templates
-from pyverilator import PyVerilator
 
 
 class HLSCustomOp(CustomOp):
@@ -33,8 +32,23 @@ class HLSCustomOp(CustomOp):
             "executable_path": ("s", False, ""),
             "ipgen_path": ("s", False, ""),
             "sim_mode": ("s", False, ""),
+            "sim_cycles": ("i", False, 0),
         }
 
+    def node_res_estimation(self):
+        resources = []
+        resources.append("BRAMs: " + str(self.bram_estimation()))
+        resources.append("LUTs: " + str(self.lut_estimation()))
+        return resources
+
+    @abstractmethod
+    def bram_estimation(self):
+        pass
+
+    @abstractmethod
+    def lut_estimation(self):
+        pass
+
     def code_generation_ipgen(self, model, fpgapart, clk):
         node = self.onnx_node
 
@@ -168,51 +182,61 @@ compilation transformations?
         process_execute.communicate()
 
     def reset_rtlsim(self, sim):
-        for i in range(10):
-            sim.io.ap_rst_n = 0
-            sim.io.ap_clk = 1
-            sim.io.ap_clk = 0
-            sim.io.ap_clk = 1
-            sim.io.ap_clk = 0
-            sim.io.ap_clk = 1
-            sim.io.ap_clk = 0
-            sim.io.ap_clk = 1
-            sim.io.ap_clk = 0
-            sim.io.ap_clk = 1
-            sim.io.ap_clk = 0
-            sim.io.ap_rst_n = 1
+        sim.io.ap_rst_n = 0
+        sim.io.ap_clk = 1
+        sim.io.ap_clk = 0
+        sim.io.ap_rst_n = 1
 
     def toggle_clk(self, sim):
-        for i in range(10):
-            sim.io.ap_clk = 1
-            sim.io.ap_clk = 0
+        sim.io.ap_clk = 1
+        sim.io.ap_clk = 0
 
     def rtlsim(self, sim, inp):
-        my_inputs = inp
-        print("My inputs before:" + str(my_inputs))
-        my_outputs = []
+        # import pdb; pdb.set_trace()
+        inputs = inp
+        outputs = []
         sim.io.out_V_V_TREADY = 1
-        for i in range(100):
-            sim.io.in0_V_V_TVALID = 1 if len(my_inputs) > 0 else 0
-            if sim.io.in0_V_V_TREADY == 1 and len(my_inputs) > 0:
-                print("ready to write input")
-                sim.io.in0_V_V_TDATA = my_inputs[0]
-                my_inputs = my_inputs[1:]
-                sim.io.ap_clk = 1
-                sim.io.ap_clk = 0
-                sim.io.in0_V_V_TVALID = 1 if len(my_inputs) > 0 else 0
-            if sim.io.out_V_V_TVALID == 1:
-                print("ready to pop result")
-                my_outputs = my_outputs + [sim.io.out_V_V_TDATA]
-                sim.io.ap_clk = 1
-                sim.io.ap_clk = 0
+
+        # observe if output is completely calculated
+        # observation_count will contain the number of cycles the calculation ran
+        num_out_values = self.get_number_output_values()
+        output_observed = False
+        observation_count = 0
+
+        # avoid infinite looping of simulation by aborting when there is no change in
+        # output values after 100 cycles
+        no_change_count = 0
+        old_outputs = outputs
+
+        while not (output_observed):
+            sim.io.in0_V_V_TVALID = 1 if len(inputs) > 0 else 0
+            sim.io.in0_V_V_TDATA = inputs[0] if len(inputs) > 0 else 0
+            if sim.io.in0_V_V_TREADY == 1 and sim.io.in0_V_V_TVALID == 1:
+                inputs = inputs[1:]
+            if sim.io.out_V_V_TVALID == 1 and sim.io.out_V_V_TREADY == 1:
+                outputs = outputs + [sim.io.out_V_V_TDATA]
             sim.io.ap_clk = 1
             sim.io.ap_clk = 0
-            print("Iteration %d" % i)
-            print(sim.io)
-            print(my_inputs)
-            print(my_outputs)
-        return my_outputs
+
+            observation_count = observation_count + 1
+            no_change_count = no_change_count + 1
+
+            if len(outputs) == num_out_values:
+                self.set_nodeattr("sim_cycles", observation_count)
+                output_observed = True
+
+            if no_change_count == 100:
+                if old_outputs == outputs:
+                    raise Exception(
+                        "Error in simulation! Takes too long to produce output."
+                    )
+                else:
+                    no_change_count = 0
+                    old_outputs = outputs
+            print(inputs)
+            print(outputs)
+
+        return outputs
 
     def execute_node(self, context, graph):
         mode = self.get_nodeattr("sim_mode")
@@ -237,6 +261,10 @@ compilation transformations?
     def generate_params(self, model, path):
         pass
 
+    @abstractmethod
+    def get_number_output_values(self):
+        pass
+
     @abstractmethod
     def global_includes(self):
         pass
diff --git a/src/finn/custom_op/fpgadataflow/convolutioninputgenerator.py b/src/finn/custom_op/fpgadataflow/convolutioninputgenerator.py
index 91460fdac4f072b0954da1152fa578cd3338fd80..60e387978a7dfd1f2ba0bff030b13d0e1f60518b 100644
--- a/src/finn/custom_op/fpgadataflow/convolutioninputgenerator.py
+++ b/src/finn/custom_op/fpgadataflow/convolutioninputgenerator.py
@@ -36,6 +36,12 @@ class ConvolutionInputGenerator(HLSCustomOp):
     def verify_node(self):
         pass
 
+    def bram_estimation(self):
+        pass
+
+    def lut_estimation(self):
+        pass
+
     def get_input_datatype(self):
         return DataType[self.get_nodeattr("inputDataType")]
 
@@ -45,6 +51,14 @@ class ConvolutionInputGenerator(HLSCustomOp):
     def get_stream_width(self):
         return self.get_nodeattr("SIMD") * self.get_nodeattr("Input_precision")
 
+    def get_number_output_values(self):
+        k = self.get_nodeattr("ConvKernelDim")
+        ifm_ch = self.get_nodeattr("IFMChannels")
+        ofm_dim = self.get_nodeattr("OFMDim")
+        out_pix = ofm_dim * ofm_dim
+
+        return out_pix * k * k * ifm_ch
+
     def execute_node(self, context, graph):
         mode = self.get_nodeattr("sim_mode")
         node = self.onnx_node
@@ -90,8 +104,15 @@ class ConvolutionInputGenerator(HLSCustomOp):
             )
             if os.path.isfile(verilog_file):
                 inp = context[node.input[0]]
-                print(inp)
+                inp = inp.transpose(0, 2, 3, 1)
                 inp = inp.flatten()
+
+                # TODO: check how to sort inputs for multichannel inputs
+                # a = []
+                # for i in range(len(inp)):
+                #     if (i+1) % 2 == 0:
+                #         a.append((int(inp[i-1]) << 1) + int(inp[i]))
+                # inp = a
                 sim = PyVerilator.build(
                     verilog_file,
                     verilog_path=[
@@ -104,12 +125,24 @@ class ConvolutionInputGenerator(HLSCustomOp):
                 super().toggle_clk(sim)
                 output = self.rtlsim(sim, inp)
                 output = [int(x) for x in output]
-                # reshape output (Only valid for sliding window!)
-                output = np.asarray(output, dtype=np.float32).reshape(
+                odt = self.get_output_datatype()
+                if odt == DataType.BIPOLAR:
+                    output = [2 * x - 1 for x in output]
+
+                # pyverilator interprets int2 as uint2, so output has to be corrected
+                elif odt == DataType.INT2:
+                    mask = 2 ** (odt.bitwidth() - 1)
+                    output = [-(x & mask) + (x & ~mask) for x in output]
+                # TODO: check how to sort inputs for multichannel inputs
+                # output = [bin(x)[2:].zfill(ifm_ch) for x in output]
+                # output_ch1 = [int(x[:1]) for x in output]
+                # output_ch2 = [int(x[1:]) for x in output]
+
+                # reshape output
+                output = np.asarray([output], dtype=np.float32).reshape(
                     1, out_pix, k * k * ifm_ch
                 )
                 context[node.output[0]] = output
-                print(output)
 
             else:
                 raise Exception(
diff --git a/src/finn/custom_op/fpgadataflow/streamingmaxpool_batch.py b/src/finn/custom_op/fpgadataflow/streamingmaxpool_batch.py
index a316695b3d8691ab66cdef7a87093d7a777ef7ff..32fba4c219886e1f1860c1a5b4d316b1fb7d8558 100644
--- a/src/finn/custom_op/fpgadataflow/streamingmaxpool_batch.py
+++ b/src/finn/custom_op/fpgadataflow/streamingmaxpool_batch.py
@@ -69,6 +69,16 @@ class StreamingMaxPool_Batch(HLSCustomOp):
 
         return info_messages
 
+
+    def get_number_output_values(self):
+        pass
+
+    def bram_estimation(self):
+        pass
+
+    def lut_estimation(self):
+        pass
+
     def global_includes(self):
         self.code_gen_dict["$GLOBALS$"] = ['#include "maxpool.h"']
 
diff --git a/src/finn/data/cpp/npy2apintstream.hpp b/src/finn/data/cpp/npy2apintstream.hpp
index f58566fb1783bbdf1e0cdbb2f69c6bd17d916e57..c058625e7c6aa1e319086db214319e0a615343c7 100644
--- a/src/finn/data/cpp/npy2apintstream.hpp
+++ b/src/finn/data/cpp/npy2apintstream.hpp
@@ -48,11 +48,11 @@ void apintstream2npy(hls::stream<PackedT> & in_stream, const std::vector<size_t>
     outer_dim_elems *= shape[dim];
   }
   size_t inner_dim_elems = shape[shape.size()-1];
-  DEBUG_NPY2APINTSTREAM("n_outer " << outer_dim_elems << " n_inner " << inner_dim_elems)
+  DEBUG_APINTSTREAM2NPY("n_outer " << outer_dim_elems << " n_inner " << inner_dim_elems)
   for(size_t outer_elem = 0; outer_elem < outer_dim_elems; outer_elem++) {
     PackedT packed_elem;
     in_stream >> packed_elem;
-    DEBUG_NPY2APINTSTREAM("packed hls elem " << std::hex << packed_elem << std::dec)
+    DEBUG_APINTSTREAM2NPY("packed hls elem " << std::hex << packed_elem << std::dec)
     for(size_t i = 0; i < inner_dim_elems; i++) {
       ElemT elem = packed_elem((i+1)*ElemBits-1, i*ElemBits);
       NpyT npyt = (NpyT) elem;
diff --git a/tests/fpgadataflow/test_fpgadataflow_convinputgenerator.py b/tests/fpgadataflow/test_fpgadataflow_convinputgenerator.py
index e32d8b765253a054d90e9c47d0e9d94202b2003b..e558807d7936e084a3cd8d8739fc4335f6642c0e 100644
--- a/tests/fpgadataflow/test_fpgadataflow_convinputgenerator.py
+++ b/tests/fpgadataflow/test_fpgadataflow_convinputgenerator.py
@@ -134,12 +134,11 @@ def prepare_inputs(input_tensor, idt):
 # input dimension
 @pytest.mark.parametrize("ifm_dim", [4, 6, 8])
 # input channels
-@pytest.mark.parametrize("ifm_ch", [1, 2, 3, 4])
+@pytest.mark.parametrize("ifm_ch", [1])  # , 2, 3, 4])
 # Stride
 @pytest.mark.parametrize("stride", [1, 2])
 def test_fpgadataflow_slidingwindow(idt, k, ifm_dim, ifm_ch, stride):
     simd = ifm_ch
-
     ofm_dim = int(((ifm_dim - k) / stride) + 1)
 
     x = gen_finn_dt_tensor(idt, (1, ifm_ch, ifm_dim, ifm_dim))
@@ -160,10 +159,12 @@ def test_fpgadataflow_slidingwindow(idt, k, ifm_dim, ifm_ch, stride):
     oshape = y_produced.shape
     y_expected = y_expected.reshape(oshape)
 
-    assert (y_produced == y_expected).all()
+    assert (y_produced == y_expected).all(), "npysim failed"
 
+    model = model.transform(SetSimMode("rtlsim"))
     model = model.transform(GiveUniqueNodeNames())
     model = model.transform(CodeGen_ipgen("xc7z020clg400-1", 5))
     model = model.transform(HLSSynth_IPGen())
-    model = model.transform(SetSimMode("rtlsim"))
+    y_produced = oxe.execute_onnx(model, input_dict)["outp"]
+    assert (y_produced == y_expected).all(), "rtlsim failed"
     model = model.transform(CleanUp())
diff --git a/tests/fpgadataflow/test_fpgadataflow_fclayer.py b/tests/fpgadataflow/test_fpgadataflow_fclayer.py
index c57b2734680319557741db7b0d49c1d6aa6d15aa..0c40000762019c93c049eeadc684c5f6043f6fcb 100644
--- a/tests/fpgadataflow/test_fpgadataflow_fclayer.py
+++ b/tests/fpgadataflow/test_fpgadataflow_fclayer.py
@@ -5,6 +5,7 @@ from onnx import TensorProto, helper
 
 import finn.core.onnx_exec as oxe
 import finn.custom_op.xnorpopcount as xp
+from finn.analysis.fpgadataflow.hls_synth_res_estimation import hls_synth_res_estimation
 from finn.core.datatype import DataType
 from finn.core.modelwrapper import ModelWrapper
 from finn.core.utils import calculate_signed_dot_prod_range, gen_finn_dt_tensor
@@ -14,6 +15,7 @@ from finn.transformation.fpgadataflow.codegen_ipgen import CodeGen_ipgen
 from finn.transformation.fpgadataflow.codegen_npysim import CodeGen_npysim
 from finn.transformation.fpgadataflow.compile import Compile
 from finn.transformation.fpgadataflow.hlssynth_ipgen import HLSSynth_IPGen
+from finn.transformation.fpgadataflow.set_sim_mode import SetSimMode
 from finn.transformation.general import GiveUniqueNodeNames
 
 
@@ -150,6 +152,7 @@ def test_fpgadataflow_fclayer(idt, wdt, act, nf, sf, mw, mh):
         else:
             tdt = DataType.INT32
     model = make_single_fclayer_modelwrapper(W, pe, simd, wdt, idt, odt, T, tdt)
+    model = model.transform(SetSimMode("npysim"))
     model = model.transform(CodeGen_npysim())
     model = model.transform(Compile())
     # prepare input data
@@ -171,8 +174,17 @@ def test_fpgadataflow_fclayer(idt, wdt, act, nf, sf, mw, mh):
     y_expected = y.reshape(oshape)
     # execute model
     y_produced = oxe.execute_onnx(model, input_dict)["outp"]
-    assert (y_produced.reshape(y_expected.shape) == y_expected).all()
+    assert (y_produced.reshape(y_expected.shape) == y_expected).all(), "npysim failed"
+    # TODO split up into several dependent tests -- need to check how this
+    # works for parametrized tests...
+    model = model.transform(SetSimMode("rtlsim"))
     model = model.transform(GiveUniqueNodeNames())
     model = model.transform(CodeGen_ipgen("xc7z020clg400-1", 5))
     model = model.transform(HLSSynth_IPGen())
+    y_produced = oxe.execute_onnx(model, input_dict)["outp"]
+    assert (y_produced.reshape(y_expected.shape) == y_expected).all(), "rtlsim failed"
+
+    hls_synt_res_est = model.analysis(hls_synth_res_estimation)
+    assert "StreamingFCLayer_Batch_0" in hls_synt_res_est
+
     model = model.transform(CleanUp())
diff --git a/tests/fpgadataflow/test_fpgadataflow_res_estimate.py b/tests/fpgadataflow/test_fpgadataflow_res_estimate.py
new file mode 100644
index 0000000000000000000000000000000000000000..50b853bf9f1aa4a7858ee9c56a79494d92fee499
--- /dev/null
+++ b/tests/fpgadataflow/test_fpgadataflow_res_estimate.py
@@ -0,0 +1,73 @@
+from onnx import TensorProto, helper
+
+from finn.analysis.fpgadataflow.res_estimation import res_estimation
+from finn.core.datatype import DataType
+from finn.core.modelwrapper import ModelWrapper
+from finn.transformation.general import GiveUniqueNodeNames
+
+
+def check_two_dict_for_equality(dict1, dict2):
+    for key in dict1:
+        assert key in dict2, "Key: {} is not in both dictionaries".format(key)
+        assert (
+            dict1[key] == dict2[key]
+        ), """Values for key {} are not the same
+        in both dictionaries""".format(
+            key
+        )
+
+    return True
+
+
+def test_res_estimate():
+    mw = mh = 4
+    simd = 1
+    pe = 1
+    idt = DataType.INT2
+    wdt = DataType.INT2
+    odt = DataType.INT32
+    actval = odt.min()
+
+    inp = helper.make_tensor_value_info("inp", TensorProto.FLOAT, [1, mw])
+    outp = helper.make_tensor_value_info("outp", TensorProto.FLOAT, [1, mh])
+    node_inp_list = ["inp", "weights", "thresh"]
+
+    FCLayer_node = helper.make_node(
+        "StreamingFCLayer_Batch",
+        node_inp_list,
+        ["outp"],
+        domain="finn",
+        backend="fpgadataflow",
+        resType="ap_resource_lut()",
+        MW=mw,
+        MH=mh,
+        SIMD=simd,
+        PE=pe,
+        inputDataType=idt.name,
+        weightDataType=wdt.name,
+        outputDataType=odt.name,
+        ActVal=actval,
+        binaryXnorMode=0,
+        noActivation=0,
+    )
+    graph = helper.make_graph(
+        nodes=[FCLayer_node], name="fclayer_graph", inputs=[inp], outputs=[outp]
+    )
+
+    model = helper.make_model(graph, producer_name="fclayer-model")
+    model = ModelWrapper(model)
+
+    model.set_tensor_datatype("inp", idt)
+    model.set_tensor_datatype("outp", odt)
+    model.set_tensor_datatype("weights", wdt)
+
+    model = model.transform(GiveUniqueNodeNames())
+    prod_resource_estimation = model.analysis(res_estimation)
+    expect_resource_estimation = {
+        "StreamingFCLayer_Batch_0": ["BRAMs: 1", "LUTs: 304.4"]
+    }
+
+    assert check_two_dict_for_equality(
+        prod_resource_estimation, expect_resource_estimation
+    ), """The produced output of
+    the resource estimation analysis pass is not equal to the expected one"""
diff --git a/tests/fpgadataflow/test_rtlsim2npy.py b/tests/fpgadataflow/test_rtlsim2npy.py
new file mode 100644
index 0000000000000000000000000000000000000000..9c8087906dd9e5700811568fb3459cc1c168ac32
--- /dev/null
+++ b/tests/fpgadataflow/test_rtlsim2npy.py
@@ -0,0 +1,50 @@
+import numpy as np
+
+from finn.core.datatype import DataType
+from finn.core.utils import unpack_innermost_dim_from_hex_string
+
+
+def test_unpack_innermost_dim_from_hex_string():
+    A = np.asarray(["0x0e", "0x06"])
+    A = A.flatten()
+    A = list(A)
+    dtype = DataType.BINARY
+    shape = (1, 2, 4)
+    packedBits = 8
+    targetBits = 1
+    eA = [[1, 1, 1, 0], [0, 1, 1, 0]]
+    A_unpacked = unpack_innermost_dim_from_hex_string(
+        A, dtype, shape, packedBits, targetBits
+    )
+    assert (A_unpacked == eA).all()
+
+    A = np.asarray(["0x0e", "0x06"])
+    A = A.flatten()
+    A = list(A)
+    eA_flipped = [[0, 1, 1, 1], [0, 1, 1, 0]]
+    A_unpacked_flipped = unpack_innermost_dim_from_hex_string(
+        A, dtype, shape, packedBits, targetBits, True
+    )
+    assert (A_unpacked_flipped == eA_flipped).all()
+
+    B = np.asarray([["0x0f", "0x0f"], ["0x07", "0x0d"]])
+    B = B.flatten()
+    B = list(B)
+    dtype = DataType.UINT2
+    shape = (1, 2, 2, 2)
+    packedBits = 8
+    targetBits = 2
+    eB = [[[3, 3], [3, 3]], [[1, 3], [3, 1]]]
+    B_unpacked = unpack_innermost_dim_from_hex_string(
+        B, dtype, shape, packedBits, targetBits
+    )
+    assert (B_unpacked == eB).all()
+
+    B = np.asarray([["0x0f", "0x0f"], ["0x07", "0x0d"]])
+    B = B.flatten()
+    B = list(B)
+    eB_flipped = [[[3, 3], [3, 3]], [[3, 1], [1, 3]]]
+    B_unpacked_flipped = unpack_innermost_dim_from_hex_string(
+        B, dtype, shape, packedBits, targetBits, True
+    )
+    assert (B_unpacked_flipped == eB_flipped).all()
diff --git a/src/finn/custom_op/fpgadataflow/streamingfclayer_batch.py b/src/finn/custom_op/fpgadataflow/streamingfclayer_batch.py
index 8d3d063c7a35bb86c72f1fcc1886c1821dd10c6a..acd577be3d5d58d1abcef8a1a2b64a029e9ced62 100644
--- a/src/finn/custom_op/fpgadataflow/streamingfclayer_batch.py
+++ b/src/finn/custom_op/fpgadataflow/streamingfclayer_batch.py
@@ -1,8 +1,14 @@
+import math
 import os
 
 import numpy as np
+from pyverilator import PyVerilator
 
-from finn.backend.fpgadataflow.utils import numpy_to_hls_code
+from finn.backend.fpgadataflow.utils import (
+    npy_to_rtlsim_input,
+    numpy_to_hls_code,
+    rtlsim_output_to_npy,
+)
 from finn.core.datatype import DataType
 from finn.core.utils import interleave_matrix_outer_dim_from_partitions
 from finn.custom_op.fpgadataflow import HLSCustomOp
@@ -141,6 +147,44 @@ class StreamingFCLayer_Batch(HLSCustomOp):
 
         return info_messages
 
+    def bram_estimation(self):
+        """the calculations are based on:
+        - FINN-R: An End-to-End Deep-Learning Framework for Fast
+        Exploration of Quantized Neural Networks
+        - M. Blott, T. B. Preusser, N. J. Fraser, G. Gambardella, K. O'Brien,
+        Y. Umuroglu, M. Leeser and K. Vissers
+        - 12. Sep 2018
+        """
+        P = self.get_nodeattr("PE")
+        Q = self.get_nodeattr("SIMD")
+        wdt = self.get_weight_datatype()
+        W = wdt.bitwidth()
+        D_in = self.get_instream_width()
+        D_out = self.get_outstream_width()
+        omega = (D_in * D_out) / (Q * P)
+        return P * (math.ceil(omega / 512)) * (math.ceil((Q * W) / 36))
+
+    def lut_estimation(self):
+        """the calculations are based on:
+        - FINN-R: An End-to-End Deep-Learning Framework for Fast
+        Exploration of Quantized Neural Networks
+        - M. Blott, T. B. Preusser, N. J. Fraser, G. Gambardella, K. O'Brien,
+        Y. Umuroglu, M. Leeser and K. Vissers
+        - 12. Sep 2018
+        """
+        P = self.get_nodeattr("PE")
+        Q = self.get_nodeattr("SIMD")
+        wdt = self.get_weight_datatype()
+        W = wdt.bitwidth()
+        # determine tdt with input and weight data types
+        idt = self.get_input_datatype()
+        A = idt.bitwidth()
+        # parameters from experiments in paper mentioned above
+        c0 = 300
+        c1 = 1.1
+
+        return c0 + c1 * (P * Q) * (W * A)
+
     def get_input_datatype(self):
         return DataType[self.get_nodeattr("inputDataType")]
 
@@ -158,6 +202,11 @@ class StreamingFCLayer_Batch(HLSCustomOp):
         o_bits = self.get_output_datatype().bitwidth()
         return o_bits * self.get_nodeattr("PE")
 
+    def get_number_output_values(self):
+        mh = self.get_nodeattr("MH")
+        pe = self.get_nodeattr("PE")
+        return mh // pe
+
     def get_template_param_values(self):
         ret = dict()
         inp_hls_str = self.get_input_datatype().get_hls_datatype_str()
@@ -347,6 +396,7 @@ class StreamingFCLayer_Batch(HLSCustomOp):
                 f_thresh.close()
 
     def execute_node(self, context, graph):
+        mode = self.get_nodeattr("sim_mode")
         node = self.onnx_node
         mw = self.get_nodeattr("MW")
         mh = self.get_nodeattr("MH")
@@ -356,7 +406,18 @@ class StreamingFCLayer_Batch(HLSCustomOp):
         nf = mh // pe
 
         # TODO ensure codegen dir exists
-        code_gen_dir = self.get_nodeattr("code_gen_dir_npysim")
+        if mode == "npysim":
+            code_gen_dir = self.get_nodeattr("code_gen_dir_npysim")
+        elif mode == "rtlsim":
+            code_gen_dir = self.get_nodeattr("code_gen_dir_ipgen")
+        else:
+            raise Exception(
+                """Invalid value for attribute sim_mode! Is currently set to: {}
+            has to be set to one of the following value ("npysim", "rtlsim")""".format(
+                    mode
+                )
+            )
+
         # create a npy file fore each input of the node (in_ind is input index)
         in_ind = 0
         for inputs in node.input:
@@ -373,6 +434,9 @@ class StreamingFCLayer_Batch(HLSCustomOp):
                 if self.get_input_datatype() == DataType.BIPOLAR:
                     # store bipolar activations as binary
                     reshaped_input = (reshaped_input + 1) / 2
+                    export_idt = DataType.BINARY
+                else:
+                    export_idt = self.get_input_datatype()
                 np.save(
                     os.path.join(code_gen_dir, "input_{}.npy".format(in_ind)),
                     reshaped_input,
@@ -380,18 +444,67 @@ class StreamingFCLayer_Batch(HLSCustomOp):
             elif in_ind > 2:
                 raise Exception("Unexpected input found for StreamingFCLayer")
             in_ind += 1
-        # execute the precompiled model
-        super().exec_precompiled_singlenode_model()
-        # load output npy file
-        super().npy_to_dynamic_output(context)
-        # reinterpret binary output as bipolar where needed
-        if self.get_output_datatype() == DataType.BIPOLAR:
-            out = context[node.output[0]]
-            out = 2 * out - 1
-            context[node.output[0]] = out
-        assert context[node.output[0]].shape == (1, nf, pe)
-        # reshape output to have expected shape
-        context[node.output[0]] = context[node.output[0]].reshape(1, mh)
+
+        if mode == "npysim":
+            # execute the precompiled model
+            super().exec_precompiled_singlenode_model()
+            # load output npy file
+            super().npy_to_dynamic_output(context)
+            # reinterpret binary output as bipolar where needed
+            if self.get_output_datatype() == DataType.BIPOLAR:
+                out = context[node.output[0]]
+                out = 2 * out - 1
+                context[node.output[0]] = out
+            assert context[node.output[0]].shape == (1, nf, pe)
+            # reshape output to have expected shape
+            context[node.output[0]] = context[node.output[0]].reshape(1, mh)
+        elif mode == "rtlsim":
+            # check if needed file exists
+            verilog_file = "{}/project_{}/sol1/impl/verilog/{}.v".format(
+                code_gen_dir, node.name, node.name
+            )
+            if os.path.isfile(verilog_file):
+                nbits = self.get_instream_width()
+                inp = npy_to_rtlsim_input(
+                    "{}/input_0.npy".format(code_gen_dir), export_idt, nbits
+                )
+                sim = PyVerilator.build(
+                    verilog_file,
+                    verilog_path=[
+                        "{}/project_{}/sol1/impl/verilog/".format(
+                            code_gen_dir, node.name
+                        )
+                    ],
+                )
+                super().reset_rtlsim(sim)
+                super().toggle_clk(sim)
+                output = self.rtlsim(sim, inp)
+                odt = self.get_output_datatype()
+                target_bits = odt.bitwidth()
+                packed_bits = self.get_outstream_width()
+                out_npy_path = "{}/output.npy".format(code_gen_dir)
+                rtlsim_output_to_npy(
+                    output, out_npy_path, odt, (1, nf, pe), packed_bits, target_bits
+                )
+
+                # load and reshape output
+                output = np.load(out_npy_path)
+                output = np.asarray([output], dtype=np.float32).reshape(1, mh)
+                context[node.output[0]] = output
+
+            else:
+                raise Exception(
+                    """Found no verilog files for this node,
+                    did you run the codegen_ipgen transformation?"""
+                )
+
+        else:
+            raise Exception(
+                """Invalid value for attribute sim_mode! Is currently set to: {}
+            has to be set to one of the following value ("npysim", "rtlsim")""".format(
+                    mode
+                )
+            )
 
     def global_includes(self):
         self.code_gen_dict["$GLOBALS$"] = ['#include "weights.hpp"']